<?php
// file       : streetosm.php
// version    : 0.04 20111108
// history    : 
//              0.04 output of ref information before street and address information
//                   set license information
//              0.03 read configuration parameter from commandline or $_POST or $_GET
//              0.02 add HTML output
//              0.01 first version
// author     : Georg Verweyen
//              georg (at) familieverweyen.de
//              www.familieverweyen.de
// license    : CC BY-SA see http://creativecommons.org/licenses/by-sa/3.0/legalcode
// description: reads the .osm files of OSM (OpenStreetMap.org)
//              and print a list of the found streets
//              first parameter value is default value
//              work            CHECK|MAP
//              outputtyp       HTML|TEXT
//              file            <filename of an osm xml file>
//              commandline: sequence of parameter without any meaning
//              in this version work=MAP doesn't work!
//
$mtime = explode(" ",microtime());
$timebegin = $mtime[1] + $mtime[0];
error_reporting(E_ERROR | E_WARNING | E_PARSE);
//
// begin configuration part 
//
define (OUTPUTTYP, 'HTML'); // value 'TEXT' or 'HTML'
define (WORKOUTPUT, 'CHECK'); // value 'CHECK' or 'MAP'
define (PART_X, 5);         // number of pieces on the x acis (3 <= value <= 9)
//
// end configuration part
$cnt_street_tag =0;
$cnt_addr_tag   =0;
$cnt_ref_tag    =0;
$cnt_waynode    =0;
$minlat         =180;
$minlon         =180;
$maxlat         =-180;
$maxlon         =-180;
$arr_street     =array();
$arr_streetcoord=array();
$arr_str_osmid  =array();
$arr_node       =array();
$arr_waynode    =array();
$currid         ='';
$currtyp        ='';
$file='';
// default parameter
$output=OUTPUTTYP;
$part_x=PART_X;
$work  =WORKOUTPUT;
//
if (count($argv)>1) {
  // read command line arguments
  for ($i=1;$i<count($argv);$i++) {
    $tmp=strtoupper($argv[$i]);
    if (($tmp=='HTML') | ($tmp=='TEXT')) {
      $output=$tmp;
    } elseif (($tmp=='MAP') | ($tmp=='CHECK')){
      $work=$tmp;
    } elseif (is_numeric($tmp)) {
      $part_x=$tmp;
    } else {
      $file=$argv[$i];
    }
  }
} else {
  // read $_POST or $_GET arguments
  $file=get_parameter('file','');
  $output=strtoupper(get_parameter('output',''));
  if (($output != 'HTML') & ($output != 'TEXT')) {
    $output=OUTPUTTYP;
  }
  $work  =strtoupper(get_parameter('work',''));
  if (($work != 'MAP') & ($work != 'CHECK')) {
    $work=WORKOUTPUT;
  }
  $part_x=get_parameter('part_x','');
  if (is_numeric($part_x)) {
    list($part_x)=sscanf($part_x,'%d');
    if (($part_x < 3) | ($part_x > 9)) {
      $part_x=PART_X;
    }
  } else {
    $part_x=PART_X;
  }
}
// arguments was read
if ($output=='TEXT') {
  if (php_sapi_name() == 'cli') {
    $charreturn="\n"; // please don't change the string start character " to '
    $blank     =' ';  // one empty character
  } else {
    $charreturn='<br />';
    $blank     ='&#8199;';  // html character for uncompressed space
  }
}
if ($work=='CHECK') {
  if ($output=='TEXT') {
    if (php_sapi_name() == 'cli') {
      echo 'Analyzing '.$file.'...'.$charreturn;
    } else {
      echo '<html><head><style>font-family: monospace</style>'.
           '<meta http-equiv="content-type" content="text/html; charset=UTF-8"></head><body>';
      echo 'Analyzing '.$file.'...'.$charreturn;
    }
  } else {  // output='HTML'
    echo '<html><head><link rel="stylesheet" type="text/css" href="streetosm.css">'.
         '<meta http-equiv="content-type" content="text/html; charset=UTF-8"></head><body>';
    //echo '<p>Analyzing '.$file." ...<br /></p>";
  }
}
$arr_street = array ();

$starttime=date('Y-m-d H:i:s');

$cntnode=0;

if (strpos($file,'.bz2')!== False) {
  if (!($fp = bzopen($file, 'r'))) {
    die("could not open XML input");
  }
} else if (strpos($file,'.gz')!== False) {
  if (!($fp = gzopen($file, 'r'))) {
    die('could not open XML input');
  }
} else {
  if (!($fp = fopen($file, "r"))) {
    die("could not open XML input");
  }
}

if ($output=='TEXT') {
  echo date('d.m.Y H:i:s').$charreturn;
}
if ($work=='CHECK') {
  if ($output!='TEXT') {
    echo '<h1>list of streets</h1>'."\n";
    echo '<table><tr><th>&nbsp;A&nbsp;</th><th>&nbsp;S&nbsp;</th><th>&nbsp;R&nbsp;'.
         '</th><th>street name</th></tr>'."\n";
  }
}
// open data file
$maxruntime=ini_get('max_execution_time');
$error_counter=0;
$line_counter=0;
$flgdetail=false; // false if work==check
while (!feof($fp)) {
  $data = trim(fgets($fp, 4096));
  $line_counter++;
  if ($maxruntime>0) {
    // runtime check for big files
    if (($line_counter % 5000)==0) {
      $mtime = explode(" ",microtime());
      $time = $mtime[1] + $mtime[0];
      if (($maxruntime-($time-$timebegin))<0.99) {
        echo '<tr><td colspan=3><img alt=\'error\' title=\'runtime warning\''.
             '  src=\'./logoerror.png\' /></td><td>', 'need ',$time-$timebegin,' from ',
             $maxruntime,' seconds<br />data do not completed readed!</td></tr>';
        break;
      } 
    }
  }
    do {// dummy while loop start
      switch ($data[1]) { // performance issue
        case 't': {
          if (strpos($data, '<tag k=')!==false) {
            $cnttag++;
            $attrs=explode($confstringchar,$data,5);
            $arr_key[$attrs[1]]=$attrs[3];
          }
          break;
	}
        case 'n': {
          if (strpos($data, '<node id=')!==false) {
            $currtyp='N';
	    analyse_xml ($data);
	    register_object_node($flgdetail);
	    break;
	  }
          if (strpos($data, '<nd ref=')!==false) {
	    analyse_xml ($data);
	    register_object_waynode($flgdetail);
	    break;
	  }
          break;
        }
        case 'w': {
          if (strpos($data, '<way id=')!==false) {
            $currtyp='W';
	    analyse_xml ($data);
	    break;
	  }
        }
        case '/': {
          if (strpos($data, '</way>')!==false) {
            // if more than one key exists for a object, invoke 
            // the routine check_key_combination
            if (count($arr_key)>1) {
              check_key_combination ();
            }
            // then delete the information for the next object
            $arr_key=array();
	    break;
	  }
          if (strpos($data, '</relation>')!==false) {
            // if more than one key exists for a object, invoke 
            // the routine check_key_combination
            if (count($arr_key)>1) {
              check_key_combination ();
            }
            // then delete the information for the next object
            $arr_key=array();
	    break;
	  }
          if (strpos($data, '</node>')!==false) {
            // if more than one key exists for a object, invoke 
            // the routine check_key_combination
            if (count($arr_key)>1) {
              check_key_combination ();
            }
            // then delete the information for the next object
            $arr_key=array();
	    break;
	  }
          if (strpos($data, '</modify>')!==false) {
	    break;
          } 
          if (strpos($data, '</create>')!==false) {
	    break;
          }
          if (strpos($data, '</delete>')!==false) {
	    break;
          } 
          if (strpos($data, '</osm>')!==false) {
	    break;
          }
        }          
        case 'm': {
          if (strpos($data, '<member type=')!==false) {
	    break;
	  }
        } 
        case 'r': {
          if (strpos($data, '<relation id=')!==false) {
            $currtyp='R';
	    //analyse_xml ($data);
	    //register_object_relation();
	    break;
          }
	}
        case '?': {
          if (strpos($data, '<?xml')!==false) {
	    break;
          }
        }
        case 'o': {
          if (strpos($data, '<osm ')!==false) {
            $typ_file='OSM';
            // Determination of the string separator
            if (strlen($data)==strlen(str_replace('"','',$data))) {
              $confstringchar='\'';
            } else {
              $confstringchar='"';
            }
	    break;
          }
        }
        case 'b': {
          if (strpos($data, '<bounds ')!==false) {
	    analyse_xml ($data);
	    break;
	  }
        } 
        default: {
          $tmp_print=true;
        }        
      } 
    } while (false); // dummy while loop end
    if ($tmp_print && (strlen($data)>0)) {
      if ($output=='TEXT') {
        echo 'can not interpret this line: \''.$data.'\')'.$charreturn;
      } else {
        echo '<tr><td colspan=3><img alt=\'error\' title=\''.
             'can not interpret this line\'  src=\'./logoerror.png\''.
             ' /></td><td>', htmlentities($data), '</td></tr>';
      }
      $error_counter++; 
      $tmp_print=false;
    }
    if ($maxruntime>0) {
      if (($line_counter % 500)==0) {
        $mtime = explode(" ",microtime());
        $time = $mtime[1] + $mtime[0];
        if (($maxruntime-($time-$timebegin))<0.3) {
          echo '<tr><td colspan=3><img alt=\'error\' title=\'runtime warning\''.
               '  src=\'./logoerror.png\' /></td><td>', 'need ',$time-$timebegin,' from ',
               $maxruntime,' seconds<br />data do not completed readed!</td></tr>';
          break;
        }
      }
    }
}
fclose($fp);
ksort($arr_street);
if ($work=='CHECK') {
  if ($output=='TEXT') {
    echo date('d.m.Y H:i:s')."  - ".$cnt_street_tag." - ".$cnt_addr_tag." - ".$cnt_ref_tag.$charreturn;
    foreach ($arr_street as $key => $val) {
      if (substr($val,2,1)!=' ') { // output ref information
        echo str_replace(' ',$blank,$val).'  '.$key.$charreturn;
      }
    }
    foreach ($arr_street as $key => $val) {
      if (substr($val,2,1)==' ') { // output street and address information
        echo str_replace(' ',$blank,$val).'  '.$key.$charreturn;
      }
    }
    echo 'from ['.$minlon.' | '.$minlat.'] to ['.$maxlon.' | '.$maxlat.']'.$charreturn;
    die(); // Output for CHECK type TEXT done
  } else {
    foreach ($arr_street as $key => $val) {
      if (substr($val,2,1)!=' ') { // output ref information
        print_html_line ($key,$val);
      }
    }
    foreach ($arr_street as $key => $val) {
      if (substr($val,2,1)==' ') { // output street and address information
        print_html_line ($key,$val);
      }
    }
  }
  echo '</table>';
  echo '<h3>legend</h3><p><b>A</b> street name found in an address<br />'.
       '<b>S</b> street name found in a highway name<br />'.
       '<b>R</b> street name found in a highway ref.</p></body></html>';
  die(); // Output for CHECK HTML done
}

function print_html_line ($key,$val) {
  echo '<tr><td class="c">';
  if (substr($val,0,1)=='A') { 
        echo 'X'; 
    }
    echo '</td><td class="c">'; 
    if (substr($val,1,1)=='S') { 
      echo 'X'; 
    }
    echo'</td><td class="c">'; 
    if (substr($val,2,1)=='R') { 
      echo 'X'; 
    }
    $current_encoding = mb_detect_encoding($key, 'auto');
    $key = iconv($current_encoding, 'UTF-8', $key);
    echo'</td><td>'.$key.'</td></tr>'."\n";
}

function register_object_node ($detail) {
  global $cnt_node;
  global $arr_node;
  global $currid;
  global $lat;
  global $lon;
  $cnt_node++;
  if ($detail) {
    $arr_node[$currid][0]=$lon;
    $arr_node[$currid][1]=$lat;
  }
}


function register_object_waynode ($detail) {
  global $cnt_waynode;
  global $arr_waynode;
  global $currid;
  global $currtyp;
  global $currref;
  $cnt_waynode++;
  if ($detail) {
    $arr_waynode[$cnt_waynode][0]=$currtyp.$currid;
    $arr_waynode[$cnt_waynode][1]=$currref;
  }
}

function analyse_xml ($data){
  global $currid;
  global $currref;
  global $currversion;
  global $currtime;
  global $lat;
  global $lon;
  global $minlat;
  global $minlon;
  global $maxlat;
  global $maxlon;
  global $confstringchar;
  
  $attrs=explode($confstringchar,$data,27);
  $key=explode(' ',$attrs[0]);
  $attrs[0]=$key[1];
  for ($i=0;$i<count($attrs);$i=$i+2) {
    $content=trim($attrs[$i]);
    switch ($content) {
      case "id=": {
	$currid=$attrs[$i+1];
	break;
      }
      case "lat=": {
	list($lat)=sscanf($attrs[$i+1],'%f');
	break;
      }
      case "lon=": {
	list($lon)=sscanf($attrs[$i+1],'%f');
	break;
      }
      case "version=": {
	$currversion=$attrs[$i+1];
	break;
      }
      case "timestamp=": {
	$currtime=$attrs[$i+1];
        break;
      }
      case "ref=": {
	$currref=$attrs[$i+1];
        break;
      }
      case "minlat=": {
 	list($minlat)=sscanf($attrs[$i+1],'%f');
 	break;
      }
      case "minlon=": {
	list($minlon)=sscanf($attrs[$i+1],'%f');
	break;
      }
      case "maxlat=": {
	list($maxlat)=sscanf($attrs[$i+1],'%f');
	break;
      }
      case "maxlon=": {
	list($maxlon)=sscanf($attrs[$i+1],'%f');
	break;
      }
    }
  }
}

function exists_in_array($v, $value) {
  // returns -1 if $v don't exists in array $value
  $work=array_search(substr($v,0,strlen($value[0])), $value);
  if ($work===false) { 
    $work=-1; 
  }
  return $work;
}  

function check_key_combination () {
  global $arr_street;
  global $arr_str_osmid;
  global $arr_str_osmtyp;
  global $arr_key;
  global $cnt_street_tag;
  global $cnt_addr_tag;
  global $cnt_ref_tag;
  global $currid;
  global $currtyp;
  //
  // information about the address
  //
  if (array_key_exists ('addr:street',$arr_key)) {
    $value=$arr_key['addr:street'];
    if (array_key_exists($value,$arr_street)){
      $arr_street[$value]='A'.substr($arr_street[$value],1,2);
    } else {
      $arr_street[$value]='A  ';
    }
    $cnt_addr_tag++;
    $arr_str_osmid [$currtyp.$currid]=$value;
  }
  //
  // information about a highway
  //
  if (array_key_exists ('highway',$arr_key)) {
    $value=$arr_key['highway'];
    if (exists_in_array($value, array('bus_stop','motorway_junction','platform'))==-1) {
      // information highway & name
      if (array_key_exists ('name',$arr_key)) {
        $value=$arr_key['name'];
        if (array_key_exists($value,$arr_street)){
          $arr_street[$value]=substr($arr_street[$value],0,1).'S'.substr($arr_street[$value],2,1);
        } else {
          $arr_street[$value]=' S ';
        }
        $cnt_street_tag++;
        $arr_str_osmid [$currtyp.$currid]=$value;
      } // end of if name
      // information highway & ref
      if (array_key_exists ('ref',$arr_key)) {
        $value=$arr_key['ref'];
        if (array_key_exists($value,$arr_street)){
          $arr_street[$value]=substr($arr_street[$value],0,2).'R';
        } else {
          $arr_street[$value]='  R';
        }
        $cnt_ref_tag++;
        $arr_str_osmid [$currtyp.$currid]=$value;
      } // end of if ref
    }
  } // end of if highway
}

function get_parameter ($parameter, $default) {
  // get parameter value (checks the parameter in the order 
  // of $_POST and $_GET. The default value is set when 
  // there is no value
  $returnvalue='';
  if (isset($_POST[$parameter])) {
    $returnvalue=$_POST[$parameter];
  } elseif (isset($_GET[$parameter])) {
    $returnvalue=$_GET[$parameter];
  }
  if ($default!='') {
    if ($returnvalue == '') {
      $returnvalue=$default;
    }
  }
  return $returnvalue;
}
